#include "OpenCL.hpp"

namespace DECS{
    namespace OpenCL{
        OpenCL::OpenCL() : error_(CL_SUCCESS), source_(cl::Program::Sources()), argument_(std::vector<cl::Buffer>())
        {
            this->setPlatform();
            this->setContext();
            this->setDevice();
            this->queue_ = cl::CommandQueue(this->context_, this->devices_.at(0), 0, &this->error_);
        }

        OpenCL::OpenCL(cl_device_type deviceType) : error_(CL_SUCCESS), source_(cl::Program::Sources())
        {
            this->setPlatform();
            this->setContext(deviceType);
            this->setDevice();
            this->queue_ = cl::CommandQueue(this->context_, this->devices_.at(0), 0, &this->error_);
        }

        OpenCL::~OpenCL()
        {
        }

        cl_int OpenCL::setPlatform()
        {
            return cl::Platform::get(&this->platforms_);
        }

        cl_int OpenCL::setContext()
        {
            cl_context_properties properties[] =
                {CL_CONTEXT_PLATFORM, (cl_context_properties)(this->platforms_[0])(), 0};
            this->context_ = cl::Context(DEFAULT_DEVICE_TYPE, properties, 0, 0, &this->error_);
            return this->error_;
        }

        cl_int OpenCL::setContext(cl_device_type deviceType)
        {
            cl_context_properties properties[] =
                {CL_CONTEXT_PLATFORM, (cl_context_properties)(this->platforms_[0])(), 0};
            this->context_ = cl::Context(deviceType, properties, 0, 0, &this->error_);
            return this->error_;
        }

        cl_int OpenCL::setDevice()
        {
            this->devices_ = this->context_.getInfo<CL_CONTEXT_DEVICES>(&this->error_);
            return this->error_;
        }

        void OpenCL::setSource(std::string& source)
        {
            this->source_.push_back(std::make_pair(source.c_str(), source.length()));
        }

        int OpenCL::setSourceFromFile(PATH srcFilePath)
        {
            IFSTREAM srcFile(srcFilePath);
            if(!srcFile){
                return 1;
            }
            std::string src;
            std::string buf;
            while(std::getline(srcFile, buf)){
                src.append(buf);
            }
            this->setSource(src);
            return 0;
        }

        cl_int OpenCL::buildProgram()
        {

            this->program_ = cl::Program(this->context_, this->source_, &this->error_);

            if(this->error_ != CL_SUCCESS){
                return this->error_;
            }
            this->program_.build(this->devices_);
            return this->error_;
        }

        cl_int OpenCL::setKernel(std::string functionName)
        {
            this->kernel_ = cl::Kernel(this->program_, functionName.c_str(), &this->error_);
            return this->error_;
        }

        cl_int OpenCL::setKernelInputArgument(cl_int index, float* argument, const int n)
        {
            cl::Buffer memArg(this->context_, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR,
                sizeof(cl_float) * n,
                argument,
                &this->error_);
            if(this->error_ != CL_SUCCESS){
                return this->error_;
            }
            if(this->argument_.size() <= index){
                this->argument_.resize(index + 1);
            }
            this->argument_.at(index) = memArg;
            return this->kernel_.setArg(index, this->argument_.at(index));
        }

        cl_int OpenCL::setKernelOutputArgument(cl_int index, const int n)
        {
            cl::Buffer memArg(this->context_, CL_MEM_WRITE_ONLY,
                sizeof(cl_float) * n,
                0,
                &this->error_);
            if(this->error_ != CL_SUCCESS){
                return this->error_;
            }
            if(this->argument_.size() <= index){
                this->argument_.resize(index + 1);
            }
            this->argument_.at(index) = memArg;
            return this->kernel_.setArg(index, this->argument_.at(index));
        }

        cl_int OpenCL::enqueueNDRange(const int n)
        {
            return this->queue_.enqueueNDRangeKernel(this->kernel_, cl::NullRange, cl::NDRange(n));
        }

        cl_int OpenCL::readBuffer(cl_int index, float* output, const int n)
        {
            return this->queue_.enqueueReadBuffer(this->argument_.at(index), CL_TRUE, 0, sizeof(float) * n, output, 0, 0);
        }
    }
}